@@ -27,7 +27,7 @@
 #include <glib.h>
 #include <glib/gi18n.h>
 
-#include <libtracker-client/tracker.h>
+#include <libtracker-client/tracker-client.h>
 #include <libtracker-common/tracker-common.h>
 
 #define ABOUT	  \
@@ -54,6 +54,8 @@ static gboolean music_files;
 static gboolean image_files;
 static gboolean video_files;
 static gboolean document_files;
+static gboolean emails;
+static gboolean contacts;
 static gboolean print_version;
 
 static GOptionEntry entries[] = {
@@ -70,7 +72,7 @@ static GOptionEntry entries[] = {
 	  NULL
 	},
 	{ "detailed", 'd', 0, G_OPTION_ARG_NONE, &detailed,
-	  N_("Show more detailed results (only applies to general search)"),
+	  N_("Show URNs for results (doesn't apply to --music-albums, --music-artists)"),
 	  NULL
 	},
 	{ "all", 'a', 0, G_OPTION_ARG_NONE, &all,
@@ -81,22 +83,22 @@ static GOptionEntry entries[] = {
 	  N_("Search for files"),
 	  NULL
 	},
-	{ "folders", 'e', 0, G_OPTION_ARG_NONE, &folders,
+	{ "folders", 's', 0, G_OPTION_ARG_NONE, &folders,
 	  N_("Search for folders"),
 	  NULL
 	},
-	{ "music-albums", 'b', 0, G_OPTION_ARG_NONE, &music_albums,
-	  N_("Search for music albums (--all has no effect on this)"),
-	  NULL
-	},
-	{ "music-artists", 's', 0, G_OPTION_ARG_NONE, &music_artists,
-	  N_("Search for music artists (--all has no effect on this) "),
-	  NULL
-	},
 	{ "music", 'm', 0, G_OPTION_ARG_NONE, &music_files,
 	  N_("Search for music files"),
 	  NULL
 	},
+	{ "music-albums", 0, 0, G_OPTION_ARG_NONE, &music_albums,
+	  N_("Search for music albums (--all has no effect on this)"),
+	  NULL
+	},
+	{ "music-artists", 0, 0, G_OPTION_ARG_NONE, &music_artists,
+	  N_("Search for music artists (--all has no effect on this) "),
+	  NULL
+	},
 	{ "images", 'i', 0, G_OPTION_ARG_NONE, &image_files,
 	  N_("Search for image files"),
 	  NULL
@@ -109,6 +111,14 @@ static GOptionEntry entries[] = {
 	  N_("Search for document files"),
 	  NULL
 	},
+	{ "emails", 'e', 0, G_OPTION_ARG_NONE, &emails,
+	  N_("Search for emails"),
+	  NULL
+	},
+	{ "contacts", 'c', 0, G_OPTION_ARG_NONE, &contacts,
+	  N_("Search for contacts"),
+	  NULL
+	},
 	{ "version", 'V', 0, G_OPTION_ARG_NONE, &print_version,
 	  N_("Print version"),
 	  NULL
@@ -152,7 +162,6 @@ get_fts_string (GStrv    search_words,
 
 	for (i = 0; i < len; i++) {
 		g_string_append (fts, search_words[i]);
-		g_string_append_c (fts, '*');
 
 		if (i < len - 1) {
 			if (use_or_operator) {
@@ -166,23 +175,252 @@ get_fts_string (GStrv    search_words,
 	return g_string_free (fts, FALSE);
 }
 
+static void
+get_contacts_foreach (gpointer value,
+		    gpointer user_data)
+{
+	gchar **data;
+	gboolean details;
+
+	data = value;
+	details = GPOINTER_TO_INT (user_data);
+
+	if (details && data[1] && *data[1]) {
+		g_print ("  %s, %s (%s)\n", data[0], data[1], data[2]);
+	} else {
+		g_print ("  %s, %s\n", data[0], data[1]);
+	}
+}
+
+static gboolean
+get_contacts_results (TrackerClient *client,
+		      const gchar   *query,
+		      gint           search_limit,
+		      gboolean       details)
+{
+	GError *error = NULL;
+	GPtrArray *results;
+
+	results = tracker_resources_sparql_query (client, query, &error);
+
+	if (error) {
+		g_printerr ("%s, %s\n",
+		            _("Could not get search results"),
+		            error->message);
+		g_error_free (error);
+
+		return FALSE;
+	}
+
+	if (!results) {
+		g_print ("%s\n",
+		         _("No contacts were found"));
+	} else {
+		g_print (g_dngettext (NULL,
+		                      "Contact: %d",
+		                      "Contacts: %d",
+		                      results->len),
+		         results->len);
+		g_print ("\n");
+
+		g_ptr_array_foreach (results,
+		                     get_contacts_foreach,
+		                     GINT_TO_POINTER (details));
+
+		if (results->len >= search_limit) {
+			show_limit_warning ();
+		}
+
+		g_ptr_array_foreach (results, (GFunc) g_strfreev, NULL);
+		g_ptr_array_free (results, TRUE);
+	}
+
+	return TRUE;
+}
+
+static gboolean
+get_contacts (TrackerClient *client,
+	      GStrv          search_terms,
+	      gboolean       show_all,
+	      gint           search_offset,
+	      gint           search_limit,
+	      gboolean       use_or_operator,
+	      gboolean       details)
+{
+	gchar *fts;
+	gchar *query;
+	gboolean success;
+
+	fts = get_fts_string (search_terms, use_or_operator);
+
+	if (fts) {
+		query = g_strdup_printf ("SELECT tracker:coalesce(nco:fullname(?contact), \"Unknown\") nco:hasEmailAddress(?contact) ?contact "
+		                         "WHERE { "
+		                         "  ?contact a nco:Contact ;"
+		                         "  fts:match \"%s\" ."
+		                         "} "
+		                         "ORDER BY ASC(nco:fullname(?contact)) ASC(nco:hasEmailAddress(?contact)) "
+		                         "OFFSET %d "
+		                         "LIMIT %d",
+		                         fts,
+		                         search_offset,
+		                         search_limit);
+	} else {
+		query = g_strdup_printf ("SELECT tracker:coalesce(nco:fullname(?contact), \"Unknown\") nco:hasEmailAddress(?contact) ?contact "
+		                         "WHERE { "
+		                         "  ?contact a nco:Contact ."
+		                         "} "
+		                         "ORDER BY ASC(nco:fullname(?contact)) ASC(nco:hasEmailAddress(?contact)) "
+		                         "OFFSET %d "
+		                         "LIMIT %d",
+		                         search_offset,
+		                         search_limit);
+	}
+
+	success = get_contacts_results (client, query, search_limit, details);
+	g_free (query);
+	g_free (fts);
+
+	return success;
+}
+
+static void
+get_emails_foreach (gpointer value,
+		    gpointer user_data)
+{
+	gchar **data;
+	gboolean details;
+
+	data = value;
+	details = GPOINTER_TO_INT (user_data);
+
+	if (details && data[1] && *data[1]) {
+		g_print ("  %s, %s (%s)\n", data[0], data[1], data[2]);
+	} else {
+		g_print ("  %s, %s\n", data[0], data[1]);
+	}
+}
+
+static gboolean
+get_emails_results (TrackerClient *client,
+		    const gchar   *query,
+		    gint           search_limit,
+		    gboolean       details)
+{
+	GError *error = NULL;
+	GPtrArray *results;
+
+	results = tracker_resources_sparql_query (client, query, &error);
+
+	if (error) {
+		g_printerr ("%s, %s\n",
+		            _("Could not get search results"),
+		            error->message);
+		g_error_free (error);
+
+		return FALSE;
+	}
+
+	if (!results) {
+		g_print ("%s\n",
+		         _("No emails were found"));
+	} else {
+		g_print (g_dngettext (NULL,
+		                      "Email: %d",
+		                      "Emails: %d",
+		                      results->len),
+		         results->len);
+		g_print ("\n");
+
+		g_ptr_array_foreach (results,
+		                     get_emails_foreach,
+		                     GINT_TO_POINTER (details));
+
+		if (results->len >= search_limit) {
+			show_limit_warning ();
+		}
+
+		g_ptr_array_foreach (results, (GFunc) g_strfreev, NULL);
+		g_ptr_array_free (results, TRUE);
+	}
+
+	return TRUE;
+}
+
+static gboolean
+get_emails (TrackerClient *client,
+	    GStrv          search_terms,
+	    gboolean       show_all,
+	    gint           search_offset,
+	    gint           search_limit,
+	    gboolean       use_or_operator,
+	    gboolean       details)
+{
+	gchar *fts;
+	gchar *query;
+	const gchar *show_all_str;
+	gboolean success;
+
+	show_all_str = show_all ? "" : "?email tracker:available true .";
+	fts = get_fts_string (search_terms, use_or_operator);
+
+	if (fts) {
+		query = g_strdup_printf ("SELECT nmo:receivedDate(?email) nmo:messageSubject(?email) nie:url(?email) "
+		                         "WHERE { "
+		                         "  ?email a nmo:Email ;"
+		                         "  fts:match \"%s\" ."
+		                         "  %s"
+		                         "} "
+		                         "ORDER BY ASC(nmo:messageSubject(?email)) ASC(nmo:receivedDate(?email))"
+		                         "OFFSET %d "
+		                         "LIMIT %d",
+		                         fts,
+		                         show_all_str,
+		                         search_offset,
+		                         search_limit);
+	} else {
+		query = g_strdup_printf ("SELECT nmo:receivedDate(?email) nmo:messageSubject(?email) nie:url(?email) "
+		                         "WHERE { "
+		                         "  ?email a nmo:Email ."
+		                         "  %s"
+		                         "} "
+		                         "ORDER BY ASC(nmo:messageSubject(?email)) ASC(nmo:receivedDate(?email))"
+		                         "OFFSET %d "
+		                         "LIMIT %d",
+		                         show_all_str,
+		                         search_offset,
+		                         search_limit);
+	}
+
+	success = get_emails_results (client, query, search_limit, details);
+	g_free (query);
+	g_free (fts);
+
+	return success;
+}
+
 static void
 get_files_foreach (gpointer value,
                    gpointer user_data)
 {
-	gchar **data = value;
+	gchar **data;
+	gboolean details;
 
-	if (data[1] && *data[1]) {
+	data = value;
+	details = GPOINTER_TO_INT (user_data);
+
+	if (details && data[1] && *data[1]) {
 		g_print ("  %s (%s)\n", data[1], data[0]);
 	} else {
-		g_print ("  %s\n", data[0]);
+		g_print ("  %s\n", data[1]);
 	}
 }
 
 static gboolean
 get_files_results (TrackerClient *client,
                    const gchar   *query,
-                   gint           search_limit)
+                   gint           search_limit,
+		   gboolean       details)
 {
 	GError *error = NULL;
 	GPtrArray *results;
@@ -211,7 +449,7 @@ get_files_results (TrackerClient *client,
 
 		g_ptr_array_foreach (results,
 		                     get_files_foreach,
-		                     NULL);
+		                     GINT_TO_POINTER (details));
 
 		if (results->len >= search_limit) {
 			show_limit_warning ();
@@ -230,7 +468,8 @@ get_document_files (TrackerClient *client,
                     gboolean       show_all,
                     gint           search_offset,
                     gint           search_limit,
-                    gboolean       use_or_operator)
+                    gboolean       use_or_operator,
+		    gboolean       details)
 {
 	gchar *fts;
 	gchar *query;
@@ -268,7 +507,7 @@ get_document_files (TrackerClient *client,
 		                         search_limit);
 	}
 
-	success = get_files_results (client, query, search_limit);
+	success = get_files_results (client, query, search_limit, details);
 	g_free (query);
 	g_free (fts);
 
@@ -281,7 +520,8 @@ get_video_files (TrackerClient *client,
                  gboolean       show_all,
                  gint           search_offset,
                  gint           search_limit,
-                 gboolean       use_or_operator)
+                 gboolean       use_or_operator,
+		 gboolean       details)
 {
 	gchar *fts;
 	gchar *query;
@@ -319,7 +559,7 @@ get_video_files (TrackerClient *client,
 		                         search_limit);
 	}
 
-	success = get_files_results (client, query, search_limit);
+	success = get_files_results (client, query, search_limit, details);
 	g_free (query);
 	g_free (fts);
 
@@ -332,7 +572,8 @@ get_image_files (TrackerClient *client,
                  gboolean       show_all,
                  gint           search_offset,
                  gint           search_limit,
-                 gboolean       use_or_operator)
+		 gboolean       use_or_operator,
+		 gboolean       details)
 {
 	gchar *fts;
 	gchar *query;
@@ -370,7 +611,7 @@ get_image_files (TrackerClient *client,
 		                         search_limit);
 	}
 
-	success = get_files_results (client, query, search_limit);
+	success = get_files_results (client, query, search_limit, details);
 	g_free (query);
 	g_free (fts);
 
@@ -383,7 +624,8 @@ get_music_files (TrackerClient *client,
                  gboolean       show_all,
                  gint           search_offset,
                  gint           search_limit,
-                 gboolean       use_or_operator)
+                 gboolean       use_or_operator,
+		 gboolean       details)
 {
 	gchar *fts;
 	gchar *query;
@@ -421,7 +663,7 @@ get_music_files (TrackerClient *client,
 		                         search_limit);
 	}
 
-	success = get_files_results (client, query, search_limit);
+	success = get_files_results (client, query, search_limit, details);
 	g_free (query);
 	g_free (fts);
 
@@ -601,7 +843,8 @@ get_files (TrackerClient *client,
            gboolean       show_all,
            gint           search_offset,
            gint           search_limit,
-           gboolean       use_or_operator)
+	   gboolean       use_or_operator,
+	   gboolean       details)
 {
 	gchar *fts;
 	gchar *query;
@@ -639,7 +882,7 @@ get_files (TrackerClient *client,
 		                         search_limit);
 	}
 
-	success = get_files_results (client, query, search_limit);
+	success = get_files_results (client, query, search_limit, details);
 	g_free (query);
 	g_free (fts);
 
@@ -652,7 +895,8 @@ get_folders (TrackerClient *client,
              gboolean       show_all,
              gint           search_offset,
              gint           search_limit,
-             gboolean       use_or_operator)
+	     gboolean       use_or_operator,
+	     gboolean       details)
 {
 	gchar *fts;
 	gchar *query;
@@ -690,7 +934,7 @@ get_folders (TrackerClient *client,
 		                         search_limit);
 	}
 
-	success = get_files_results (client, query, search_limit);
+	success = get_files_results (client, query, search_limit, details);
 	g_free (query);
 	g_free (fts);
 
@@ -702,16 +946,16 @@ get_all_by_search_foreach (gpointer value,
                            gpointer user_data)
 {
 	gchar **metadata;
-	gboolean detailed;
+	gboolean details;
 
 	metadata = value;
-	detailed = GPOINTER_TO_INT (user_data);
+	details = GPOINTER_TO_INT (user_data);
 
 	if (!metadata || !*metadata) {
 		return;
 	}
 
-	if (G_LIKELY (!detailed)) {
+	if (G_LIKELY (!details)) {
 		gchar **p;
 		gint i;
 
@@ -732,18 +976,18 @@ get_all_by_search_foreach (gpointer value,
 		if (mime_type && mime_type[0] == '\0') {
 			mime_type = NULL;
 		}
-		
+
 		if (mime_type) {
 			g_print ("  %s\n"
 			         "    %s\n"
-			         "    %s\n", 
-			         urn, 
-			         mime_type, 
+			         "    %s\n",
+			         urn,
+			         mime_type,
 			         class);
 		} else {
 			g_print ("  %s\n"
-			         "    %s\n", 
-			         urn, 
+			         "    %s\n",
+			         urn,
 			         class);
 		}
 	}
@@ -758,7 +1002,7 @@ get_all_by_search (TrackerClient *client,
                    gint           search_offset,
                    gint           search_limit,
                    gboolean       use_or_operator,
-                   gboolean       detailed_results)
+                   gboolean       details)
 {
 	GError *error = NULL;
 	GPtrArray *results;
@@ -773,7 +1017,7 @@ get_all_by_search (TrackerClient *client,
 
 	show_all_str = show_all ? "" : "?s tracker:available true . ";
 
-	if (detailed_results) {
+	if (details) {
 		query = g_strdup_printf ("SELECT tracker:coalesce (nie:url (?s), ?s) nie:mimeType (?s) ?type "
 		                         "WHERE {"
 		                         "  ?s fts:match \"%s\" ;"
@@ -828,7 +1072,7 @@ get_all_by_search (TrackerClient *client,
 
 		g_ptr_array_foreach (results,
 		                     get_all_by_search_foreach,
-		                     GINT_TO_POINTER (detailed_results));
+		                     GINT_TO_POINTER (details));
 
 		if (results->len >= search_limit) {
 			show_limit_warning ();
@@ -886,6 +1130,8 @@ main (int argc, char **argv)
 	    !image_files &&
 	    !video_files &&
 	    !document_files &&
+	    !emails &&
+	    !contacts &&
 	    !files && !folders &&
 	    !terms) {
 		gchar *help;
@@ -907,42 +1153,48 @@ main (int argc, char **argv)
 		g_thread_init (NULL);
 	}
 
-	/* Check terms are not stopwords */
 	if (terms) {
 		TrackerLanguage *language;
-		GHashTable *stop_words;
-		const gchar *stop_word_found;
+		gboolean stop_words_found;
 		gchar **p;
 
-		language = tracker_language_new (NULL);
-		stop_words = tracker_language_get_stop_words (language);
+		/* Check terms don't have additional quotes */
+		for (p = terms; *p; p++) {
+			gchar *term = *p;
+			gint end = strlen (term) - 1;
 
-		for (p = terms, stop_word_found = NULL; *p && !stop_word_found; p++) {
-			gpointer data;
-			gchar *up = g_utf8_strdown (*p, -1);
-
-			data = g_hash_table_lookup (stop_words, up);
-			if (data) {
-				stop_word_found = *p;
+			if ((term[0] == '"' && term[end] == '"') ||
+			    (term[0] == '\'' && term[end] == '\'')) {
+				/* We never have a quote JUST at the end */
+				term[0] = term[end] = ' ';
+				g_strstrip (term);
 			}
-			g_free (up);
 		}
 
-		if (stop_word_found) {
-			g_printerr (_("Search term '%s' is a stop word."),
-			            stop_word_found);
-			g_printerr ("\n\n");
-			g_printerr (_("Stop words are common words which are "
-			              "ignored during the indexing process."));
-			g_printerr ("\n");
-			g_printerr (_("This means this search term will never "
-			              "be found when matching FTS entries."));
-			g_printerr ("\n\n");
+		/* Check if terms are stopwords, and warn if so */
+		language = tracker_language_new (NULL);
+		stop_words_found = FALSE;
+		for (p = terms; *p; p++) {
+			gchar *down;
 
-			g_option_context_free (context);
-			g_object_unref (language);
+			down = g_utf8_strdown (*p, -1);
 
-			return EXIT_FAILURE;
+			if (tracker_language_is_stop_word (language, down)) {
+				g_printerr (_("Search term '%s' is a stop word."),
+				            down);
+				g_printerr ("\n");
+
+				stop_words_found = TRUE;
+			}
+
+			g_free (down);
+		}
+
+		if (stop_words_found) {
+			g_printerr (_("Stop words are common words which "
+			              "may be ignored during the indexing "
+			              "process."));
+			g_printerr ("\n\n");
 		}
 
 		g_object_unref (language);
@@ -965,7 +1217,7 @@ main (int argc, char **argv)
 	if (files) {
 		gboolean success;
 
-		success = get_files (client, terms, all, offset, limit, or_operator);
+		success = get_files (client, terms, all, offset, limit, or_operator, detailed);
 		g_object_unref (client);
 
 		return success ? EXIT_SUCCESS : EXIT_FAILURE;
@@ -974,7 +1226,7 @@ main (int argc, char **argv)
 	if (folders) {
 		gboolean success;
 
-		success = get_folders (client, terms, all, offset, limit, or_operator);
+		success = get_folders (client, terms, all, offset, limit, or_operator, detailed);
 		g_object_unref (client);
 
 		return success ? EXIT_SUCCESS : EXIT_FAILURE;
@@ -1001,7 +1253,7 @@ main (int argc, char **argv)
 	if (music_files) {
 		gboolean success;
 
-		success = get_music_files (client, terms, all, offset, limit, or_operator);
+		success = get_music_files (client, terms, all, offset, limit, or_operator, detailed);
 		g_object_unref (client);
 
 		return success ? EXIT_SUCCESS : EXIT_FAILURE;
@@ -1010,7 +1262,7 @@ main (int argc, char **argv)
 	if (image_files) {
 		gboolean success;
 
-		success = get_image_files (client, terms, all, offset, limit, or_operator);
+		success = get_image_files (client, terms, all, offset, limit, or_operator, detailed);
 		g_object_unref (client);
 
 		return success ? EXIT_SUCCESS : EXIT_FAILURE;
@@ -1019,7 +1271,7 @@ main (int argc, char **argv)
 	if (video_files) {
 		gboolean success;
 
-		success = get_video_files (client, terms, all, offset, limit, or_operator);
+		success = get_video_files (client, terms, all, offset, limit, or_operator, detailed);
 		g_object_unref (client);
 
 		return success ? EXIT_SUCCESS : EXIT_FAILURE;
@@ -1028,7 +1280,25 @@ main (int argc, char **argv)
 	if (document_files) {
 		gboolean success;
 
-		success = get_document_files (client, terms, all, offset, limit, or_operator);
+		success = get_document_files (client, terms, all, offset, limit, or_operator, detailed);
+		g_object_unref (client);
+
+		return success ? EXIT_SUCCESS : EXIT_FAILURE;
+	}
+
+	if (emails) {
+		gboolean success;
+
+		success = get_emails (client, terms, all, offset, limit, or_operator, detailed);
+		g_object_unref (client);
+
+		return success ? EXIT_SUCCESS : EXIT_FAILURE;
+	}
+
+	if (contacts) {
+		gboolean success;
+
+		success = get_contacts (client, terms, all, offset, limit, or_operator, detailed);
 		g_object_unref (client);
 
 		return success ? EXIT_SUCCESS : EXIT_FAILURE;
